"""
   Descp: Hierarchical data transfer for use in sunburst and treemaps

   Created on: 11-mar-2022

   Copyright 2022-2022 David Davó Laviña
        <ddavo@ucm.es>
"""
import pandas as pd
from millify import millify

class HierarchicalData():
    """
    * labels
    * parents
    * ids
    * values
    * customdata
    * total
    """

    def __init__(self):
        self.labels = []
        self.parents = []
        self.ids = []
        self.values = []
        self.customdata = []
        self.total = None
    
    @staticmethod
    def from_df(df: pd.DataFrame):
        self = HierarchicalData()
        self.labels = df['label'].to_list()
        self.parents = df['parent'].to_list()
        self.ids = df['id'].to_list()
        self.values = df['value'].to_list()
        if 'customData' in df.columns:
            self.customdata = df['customData'].to_list()
        self.total = df['value'].sum()
        self.total = millify(df['value'].sum(), precision=2) + '$'

        return self
    
    def to_dict(self):
        return {
            'ids': self.ids,
            'labels': self.labels,
            'parents': self.parents,
            'values': self.values,
            'customdata': self.customdata,
            'total': self.total,
        }



"""
   Descp: This class is used to represent a multiple bar figure, this means,
        stacked multibar-bar figure, and grouped multi-bar figure.

   Created on: 14-sep-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""
from typing import Dict, List
import plotly.graph_objs as go

from dao_analyzer.apps.common.presentation.charts.layout.figure.figure import Figure

class MultiBarFigure(Figure):
    STACK: int = 0
    GROUP: int = 1
    __TYPE_STACK: str = 'stack'
    __TYPE_GROUP: str = 'group'


    def __init__(self, bar_type: int) -> None:
        super().__init__()
        super().configuration.add_x_params(params={'tickangle': True})
        super().configuration.add_y_params(params={'grid': True})
        self.__bar_type: str = self.__get_type(bar_type)


    def __get_type(self, bar_type: int) -> str:
        btype: str = self.__TYPE_GROUP

        if bar_type == MultiBarFigure.STACK:
            btype = self.__TYPE_STACK

        return btype


    @staticmethod
    def get_empty_plot_data() -> Dict:
        return {
            'common': {
                'x': list(),
                'type': '-',
                'x_format': '', 
                'ordered_keys': []
            }
        }


    def get_empty_figure(self) -> Dict:
        return self.get_figure(plot_data = self.get_empty_plot_data())


    def get_figure(self, plot_data: Dict) -> Dict:
        """
        Returns the multi-bar chart filled with plot_data.
        Arguments:
            * plot_data = {
                'type_i': { Each bar type needs an entry in this dict and
                            to be registered in plot_data['common']['ordered_keys']

                    'y': A list filled with y-axis values.
                    'name': Bar type name.
                    'color': The bar's color.
                }
                'common': { Common values for all the bars
                    'x': A list type filled with values.
                    'type': The type of the x-axis values, e.g. date.
                    'x_format': Format of the x-axis values, e.g. '%b, %Y'.
                    'ordered_keys': An ordered list of the dict-keys of each
                                    bar type. It's used to plot each bar type
                                    in a specific order.
                }
            } 
        """
        bars: List = list()
        for k in plot_data['common']['ordered_keys']:
            bars.append(go.Bar(
                    x=plot_data['common']['x'], 
                    y=plot_data[k]['y'], 
                    name=plot_data[k]['name'], 
                    marker_color=plot_data[k]['color']))

        super().configuration.add_x_params(
            params={
                'tickvals': plot_data['common']['x'],
                'type': plot_data['common']['type'],
                'tickformat': plot_data['common']['x_format'],
            })

        super().configuration.enable_legend()

        layout: go.Layout = go.Layout(
            barmode=self.__bar_type,
            xaxis=super().configuration.get_x_axis_layout(),
            yaxis=super().configuration.get_y_axis_layout(),
            legend=super().configuration.get_legend(),
            shapes=super().configuration.get_shapes(),
            margin=self.configuration.get_margin(),
        )

        return {'data': bars, 'layout': layout}



"""
   Descp: Stacked serie transfer for multiple 'y' axis values.

   Created on: 3-mar-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui 
        <f.r.youssef@hotmail.com>
"""

from typing import List, Tuple

from dao_analyzer.logs import LOGS
from dao_analyzer.apps.common.business.transfers.serie import Serie

class StackedSerie():
    """
    * serie = see Serie
    * y_stack = each element of y_stack is a list of values on the serie.
    """

    def __init__(self, serie: Serie = None, y_stack: List[List] = None):
        self.serie = serie if serie else Serie()
        self.y_stack = y_stack if y_stack else list()


    def get_serie(self) -> List:
        if not self.serie:
            raise Exception(LOGS['attr_not_init']
                .format('serie', 'get_serie'))

        return self.serie.get_x()


    def get_last_serie_elem(self) -> str:
        if not self.serie:
            raise Exception(LOGS['attr_not_init']
                .format('serie', 'get_last_serie_elem'))

        return self.serie.get_last_serie_elem()


    def get_last_value(self, i_stack: int):
        if i_stack >= len(self.y_stack):
            return 0 
        return self.y_stack[i_stack][-1]

    
    def get_i_stack(self, i_stack: int) -> List:
        if i_stack >= len(self.y_stack):
            return list() 
        return self.y_stack[i_stack]

    def get_last_values(self, i_stack: int = 0, add_stacks: bool = False) -> Tuple[float, float]:
        """
        Returns the last two values (this month and prev)
        to be able to calculate the difference between them

        * Returns this, prev
        """
        if add_stacks:
            i_stack = 0

        if i_stack >= len(self.y_stack):
            return 0.0, 0.0

        y: List = self.y_stack[i_stack]

        i_1 = -1
        i_2 = -(2 % (len(y) + 1))
        op1 = y[i_1]
        op2 = y[i_2]

        if op1 and op2 and add_stacks:
            for j in range(i_stack+1, len(self.y_stack)):
                y = self.y_stack[j]
                op1 += y[i_1]
                op2 += y[i_2]
        
        return op1, op2
    
    def get_diff_last_values(self, **kwargs) -> float:
        this, prev = self.get_last_values(**kwargs)

        if this is None:
            this = 0.0

        if prev is None:
            prev = 0.0
        
        return this - prev

    def get_rel_last_values(self, **kwargs)\
    -> float:
        """
        A percentage of the diference among the last two values.
        Parameters: 
            * i_stack: indicates the stack to operate. Default 0.
            * add_stacks: indicates whether you wish to operate over all
                          the stacks, if you enable this flag, i_stack don't
                          care. Default False.
        Return:
            A float. 
        """
        this, prev = self.get_last_values(**kwargs)

        if this is None:
            this = 0.0

        if prev is None:
            prev = 0.0

        val: float = 0.0
        denominator = (this + prev)
        numerator = (this - prev)

        if denominator > 0:
            val = numerator / denominator * 100

        return val


    def get_n_stacks(self, n_stacks: int) -> List[List]:
        stacks: List[List] = list()

        for s in self.y_stack:
            stacks.append(s)
        # fill with the remainder n_stacks
        for _ in range(len(stacks), n_stacks):
            stacks.append(list())

        return stacks


    def get_stacks(self) -> List[List]:
        return self.y_stack



"""
   Descp: Class to adapt StakedSeries in a chart. This class is used to adapt
          basic metrics.

   Created on: 19-oct-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

from typing import Dict, List

import dao_analyzer.apps.common.resources.colors as Color
from dao_analyzer.apps.common.business.transfers.stacked_serie import StackedSerie
from dao_analyzer.apps.common.business.transfers.organization import OrganizationList
from dao_analyzer.apps.common.business.i_metric_adapter import IMetricAdapter
import dao_analyzer.apps.aragon.data_access.daos.metric.\
    metric_dao_factory as s_factory

class BasicAdapter(IMetricAdapter):

    DATE_FORMAT: str = '%b, %Y'

    def __init__(self, metric_id: int) -> None:
        self.__metric = metric_id

    def get_plot_data(self, o_id: str, organizations: OrganizationList) -> Dict:
        """
        Returns the metric data in a Dict using o_id param.
        """
        dao = s_factory.get_dao(
            ids=organizations.get_ids_from_id(o_id),
            metric=self.__metric
        )
        metric: StackedSerie = dao.get_metric()

        y: List[float] = metric.get_i_stack(0)
        color = [Color.LIGHT_BLUE] * len(y)

        if color:
            color[-1] = Color.DARK_BLUE

        return {
            'x': metric.get_serie(),
            'y': y,
            'name': '',
            'color': color,
            'type': 'date',
            'x_format': self.DATE_FORMAT,
            'last_serie_elem': metric.get_last_serie_elem(),
            'last_value': metric.get_last_value(0),
            'diff': metric.get_diff_last_values(),
            'diff_rel': metric.get_rel_last_values(),
        }



"""
   Descp: This class is used to represent a treemap

   Created on: 10-mar-2022

   Copyright 2020-2022 David Davó
        <ddavo@ucm.es>
"""
from typing import Dict
import plotly.graph_objects as go

from . import Figure

class TreemapFigure(Figure):
    def __init__(self) -> None:
        super().__init__()
    
    @staticmethod
    def get_empty_plot_data() -> Dict:
        return {
            'ids': [],
            'labels': [],
            'parents': [],
            'values': []
        }
    
    def get_empty_figure(self) -> Dict:
        return self.get_figure(plot_data=self.get_empty_plot_data())
    
    def get_figure(self, plot_data: Dict) -> Dict:
        """
        Returns the treemap filled with plot_data
        Arguments:
            * plot_data: see https://plotly.com/python/reference/treemap/
        """
        # TODO: Instead of popping 'total', pass only needed values

        plot_data.pop('total', None)
        treemap: go.Treemap = go.Treemap(**plot_data)
        layout: go.Layout = go.Layout(
            margin=self.configuration.get_margin(),
        )

        return {'data': [treemap], 'layout': layout}



"""
   Descp: This transfer has a stack of StackedSerie.
          Use it for example for multiple stacked bar chart.

   Created on: 3-mar-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui 
        <f.r.youssef@hotmail.com>
"""

from typing import List

from dao_analyzer.logs import LOGS
from dao_analyzer.apps.common.business.transfers.serie import Serie
from dao_analyzer.apps.common.business.transfers.stacked_serie import StackedSerie

class NStackedSerie():
    """
    NStackedSerie uses StackedSerie and Serie with composition 
    instead of inheritance for better readability and traceability.

    * serie: see Serie.
    * sseries: a list of stacked 'y' axis sseries. See StackedSerie
    """

    def __init__(self, serie: Serie = None, sseries: List[StackedSerie] = None):
        self.__serie = serie if serie else Serie()
        self.__sseries = sseries if sseries else list()


    def get_serie(self) -> List:
        if not self.__serie:
            raise Exception(LOGS['attr_not_init']
                .format('serie', 'get_serie'))

        return self.__serie.get_x()

    
    def get_last_serie_elem(self) -> str:
        if not self.__serie:
            raise Exception(LOGS['attr_not_init']
                .format('serie', 'get_last_serie_elem'))

        return self.__serie.get_last_serie_elem()


    def get_last_value(self, i_value: int, j_stack: int) -> int:
        if not self.__sseries:
            raise Exception(LOGS['attr_not_init']
                .format('sseries', 'get_last_value'))

        return self.__sseries[i_value].get_last_value(j_stack)

    
    def get_i_stack(self, i_value: int, j_stack: int) -> List[int]:
        if not self.__sseries:
            raise Exception(LOGS['attr_not_init']
                .format('sseries', 'get_i_stack'))

        return self.__sseries[i_value].get_i_stack(j_stack)


    def get_diff_last_sseries(self, i_value: int, j_stack: int) -> int:
        if not self.__sseries:
            raise Exception(LOGS['attr_not_init']
                .format('sseries', 'get_diff_last_sseries'))

        return self.__sseries[i_value].get_diff_last_sseries(j_stack)


    def get_i_sserie(self, i: int) -> StackedSerie:
        if i >= len(self.__sseries):
            return StackedSerie()

        return self.__sseries[i]


"""
   Descp: Creates a Singleton metaclass used by the app services

   Created on: 22-sep-2021

   Copyright 2021-2021 David Davó Laviña
        <david@ddavo.me>
"""
from abc import ABCMeta
import threading
import inspect

def makeHashable(o):
    if isinstance(o, tuple) or isinstance(o, set) or isinstance (o, list):
        return frozenset([makeHashable(x) for x in o])
    elif isinstance(o, dict):
        return frozenset([(k, makeHashable(v)) for k,v in o.items()])
    else:
        return o

# Extract by https://stackoverflow.com/q/6760685/4505998
# Added support for call with arguments (which return different instantiations)
class Singleton(type):
    _instances = {}
    _init = {}

    def _s_get_key(cls, args, kwargs):
        if cls in cls._init:
            return (cls, makeHashable(inspect.getcallargs(cls._init[cls], None, *args, **kwargs)))
        else:
            return cls

    def __init__(cls, name, bases, dict):
        super().__init__(name, bases, dict)
        cls._init[cls] = dict.get('__init__', None)

    def __call__(cls, *args, **kwargs):
        key = cls._s_get_key(args, kwargs)
        if key not in cls._instances:
            cls._instances[key] = super().__call__(*args, **kwargs)
        return cls._instances[key]

# Extract from https://stackoverflow.com/a/51897195/4505998
class ThreadSafeSingleton(Singleton):
    __lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        # Double check to avoid locking when reading
        key = cls._s_get_key(args, kwargs)
        if key not in cls._instances:
            with cls.__lock:
                return super().__call__(*args, **kwargs)

        return cls._instances[key]

# To avoid meta clashes
class ABCSingleton(Singleton, ABCMeta):
    pass

class ABCThreadSafeSingleton(ThreadSafeSingleton, ABCMeta):
    pass


from pathlib import Path

from . import __version__

CACHE_SCRIPTS_VERSION = __version__

# https://letsexchange.io/blog/what-is-block-confirmation-on-ethereum-and-how-many-confirmations-are-required/
# Number of blocks to skip to only consult confirmed blocks
SKIP_INVALID_BLOCKS = 250
DEFAULT_DATAWAREHOUSE = Path('datawarehouse')

__args = None

def populate_args(args):
    global __args
    __args = args


def __getattr__(name):
    """
    Called when no function has been defined. Defaults to search argsparser.
    """
    return __args.__getattribute__(name)



"""
   Descp: Common elements of the view.

   Created on: 1-oct-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""
import os
from dash import html
import dash_bootstrap_components as dbc

from dao_analyzer import __version__
from dao_analyzer.apps.common.resources.strings import TEXT

REL_PATH: str = os.path.join('..', '..', '..', '..', 'assets')


def generate_layout(body: html.Div) -> html.Div:

    return html.Div(children=[
        __generate_header(),
        body,
        __generate_foot(),
    ], className='root',)


def __generate_header() -> html.Div:
    return html.Div(children=[
        html.Div([
            html.Img(src=os.path.join(REL_PATH, TEXT['dao_analyzer_logo_name']), className='header-logo'),
            html.Span(TEXT['dao_analyzer_header_text'], className='header-text'),
        ], className='header-things-container'),
        dbc.Container([
            html.A('About', href="/about", className="header-link"),
        ], className='header-links-container'),
    ], className='main-header')


def __generate_foot() -> html.Div:
    return html.Footer(
        dbc.Container(children=[
            dbc.Row([
                html.Div([
                    html.I(className='bi bi-github'),
                    ' ',
                    TEXT['follow_us'],
                    html.A(TEXT['github'], href=TEXT['github_url'], target='_black', className='url-color')
                ], className='col'),
                html.Div(TEXT['current_version'].format(version=__version__), className='col-12 col-sm text-sm-end'),
            ], className='mb-3'),
            dbc.Row([
                html.Div([
                    html.Div(children=[
                        html.A(children=[
                            html.Img(src=os.path.join(REL_PATH, TEXT['cc_image_name']), className='license-img'),
                        ], href=TEXT['cc_url'], target='_blank'),
                        html.A(children=[
                            html.Img(src=os.path.join(REL_PATH, TEXT['gpl_image_name']), className='license-img'),
                        ], href=TEXT['gpl_url'], target='_blank'),
                    ], className='foot-ack-logos'),
                    html.Div(children=[
                        html.P([
                            f"{TEXT['cc_license_text']}. {TEXT['gpl_license_text']}. ",
                            html.A(TEXT['p2p_models'], href=TEXT['p2p_models_url'], target='_blank', className='url-color'),
                            f"{TEXT['acknowledgements']}."
                        ]),
                    ]),
                ], className='col d-flex flex-row'),
                html.Div([
                    html.A(children=[
                        html.Img(src=os.path.join(REL_PATH, TEXT['spanish_ministry_image_name']),
                            className='sponsor-img'),
                    ], href=TEXT['spanish_ministry_url'], target='_blank'),
                    html.A(children=[
                        html.Img(src=os.path.join(REL_PATH, TEXT['logo_ucm_name']),
                            className='sponsor-img'),
                    ], href=TEXT['ucm_url'], target='_blank'),
                    html.A(children=[
                        html.Img(src=os.path.join(REL_PATH, TEXT['logo_grasia_name']),
                            className='sponsor-img'),
                    ], href=TEXT['grasia_url'], target='_blank'),
                    html.A(children=[
                        html.Img(src=os.path.join(REL_PATH, TEXT['erc_image_name']),
                            className='sponsor-img'),
                    ], href=TEXT['erc_url'], target='_blank'),
                ], className='col footer-logos-container'),
            ], className='row-cols-1 row-cols-lg-2 gy-3'),
        ]), className='mt-3'
    )



"""
   Descp: This file is used as factory to create a DAO metric

   Created on: 19-oct-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""
from typing import List, Dict, Tuple
from dao_analyzer.apps.aragon.data_access.daos.metric.strategy.st_organization_activity import StOrganizationActivity
from dao_analyzer.apps.common.data_access.daos.metric.strategy import IMetricStrategy

from dao_analyzer.apps.common.data_access.daos.metric.metric_dao \
    import MetricDao
from dao_analyzer.apps.common.data_access.requesters import CacheRequester, JoinCacheRequester
import dao_analyzer.apps.aragon.data_access.daos.metric.srcs as srcs
from dao_analyzer.apps.aragon.data_access.daos.metric.strategy.st_new_additions import StNewAdditions
from dao_analyzer.apps.aragon.data_access.daos.metric.strategy.st_installed_apps import StInstalledApps
from dao_analyzer.apps.aragon.data_access.daos.metric.strategy.st_cast_type import StCastType
from dao_analyzer.apps.aragon.data_access.daos.metric.strategy.st_vote_outcome import StVoteOutcome
from dao_analyzer.apps.aragon.data_access.daos.metric.strategy.st_active_voters import StActiveVoters
from dao_analyzer.apps.aragon.data_access.daos.metric.strategy.st_active_token_holders import StActiveTokenHolders
from dao_analyzer.apps.aragon.data_access.daos.metric.strategy.st_active_organization import StActiveOrganization
from dao_analyzer.apps.aragon.data_access.daos.metric.strategy.st_approval_vote_rate import StApprovalVoteRate
from dao_analyzer.apps.aragon.data_access.daos.metric.strategy.st_casted_votes_voters_rate import StVoteVotersRate
from dao_analyzer.apps.aragon.data_access.daos.metric.strategy.st_casted_votes_rate import StCastedVotesRate
from dao_analyzer.apps.aragon.data_access.daos.metric.strategy.st_assets_tokens import StAssetsTokens
from dao_analyzer.apps.aragon.data_access.daos.metric.strategy.st_assets_values import StAssetsValues
from dao_analyzer.apps.aragon.data_access.daos.metric.strategy.st_total_token_holders import StTotalTokenHolders

NEW_VOTES = 0
NEW_TRANSACTIONS = 1
INSTALLED_APPS = 2
CAST_TYPE = 3
VOTE_OUTCOME = 4
ACTIVE_VOTERS = 5
ACTIVE_TOKEN_HOLDERS = 6
ACTIVE_ORGANIZATION = 7
APPROVAL_VOTE_RATE = 8
CASTED_VOTE_VOTER_RATE = 9
CASTED_VOTE_FOR_RATE = 10
CASTED_VOTE_AGAINST_RATE = 11
ASSETS_VALUES = 12
ASSETS_TOKENS = 13
ORGANIZATION_ACTIVITY = 14
TOTAL_TOKEN_HOLDERS = 15


def _metricsDefault(metric: int) -> Tuple[IMetricStrategy, CacheRequester]: # noqa: C901
    address_key: str = ''
    requester: CacheRequester = None
    stg = None

    if metric == NEW_VOTES:
        stg = StNewAdditions(typ=StNewAdditions.VOTE)
        requester = CacheRequester(srcs=[srcs.VOTES])
        address_key = 'orgAddress'
    elif metric == NEW_TRANSACTIONS:
        stg = StNewAdditions(typ=StNewAdditions.TRANSACTION)
        requester = CacheRequester(srcs=[srcs.TRANSACTIONS])
        address_key = 'orgAddress'
    elif metric == INSTALLED_APPS:
        stg = StInstalledApps()
        requester = CacheRequester(srcs=[srcs.APPS])
        address_key = 'organizationId'
    elif metric == CAST_TYPE:
        stg = StCastType()
        requester = CacheRequester(srcs=[srcs.CASTS])
        address_key = 'orgAddress'
    elif metric == VOTE_OUTCOME:
        stg = StVoteOutcome()
        requester = CacheRequester(srcs=[srcs.VOTES])
        address_key = 'orgAddress'
    elif metric == ACTIVE_VOTERS:
        stg = StActiveVoters()
        requester = CacheRequester(srcs=[srcs.CASTS])
        address_key = 'orgAddress'
    elif metric == ACTIVE_TOKEN_HOLDERS:
        stg = StActiveTokenHolders()
        requester = CacheRequester(srcs=[
            srcs.CASTS,
            srcs.VOTES,
            srcs.TRANSACTIONS])
        address_key = 'orgAddress'
    elif metric == ACTIVE_ORGANIZATION:
        stg = StActiveOrganization()
        requester = CacheRequester(srcs=[
            srcs.CASTS,
            srcs.VOTES,
            srcs.TRANSACTIONS])
        address_key = 'orgAddress'
    elif metric == APPROVAL_VOTE_RATE:
        stg = StApprovalVoteRate()
        requester = CacheRequester(srcs=[srcs.VOTES])
        address_key = 'orgAddress'
    elif metric == CASTED_VOTE_VOTER_RATE:
        stg = StVoteVotersRate()
        requester = CacheRequester(srcs=[srcs.CASTS])
        address_key = 'orgAddress'
    elif metric == CASTED_VOTE_FOR_RATE:
        stg = StCastedVotesRate(m_type=StCastedVotesRate.CAST_VOTE_FOR)
        requester = CacheRequester(srcs=[srcs.CASTS])
        address_key = 'orgAddress'
    elif metric == CASTED_VOTE_AGAINST_RATE:
        stg = StCastedVotesRate(m_type=StCastedVotesRate.CAST_VOTE_AGAINST)
        requester = CacheRequester(srcs=[srcs.CASTS])
        address_key = 'orgAddress'
    elif metric == ASSETS_VALUES:
        address_key = 'orgAddress'
        stg = StAssetsValues()
        requester = JoinCacheRequester(srcs=[
            srcs.ORGANIZATIONS,
            srcs.TOKEN_BALANCES
        ], on=['recoveryVault', 'network'])
    elif metric == ASSETS_TOKENS:
        address_key = 'orgAddress'
        stg = StAssetsTokens()
        requester = JoinCacheRequester(srcs=[
            srcs.ORGANIZATIONS,
            srcs.TOKEN_BALANCES
        ], on=['recoveryVault', 'network'])
    elif metric == ORGANIZATION_ACTIVITY:
        stg = StOrganizationActivity()
        requester = CacheRequester(srcs=[
            srcs.CASTS,
            srcs.VOTES,
            srcs.TRANSACTIONS])
        address_key = 'orgAddress'
    elif metric == TOTAL_TOKEN_HOLDERS:
        stg = StTotalTokenHolders()
        requester = CacheRequester(srcs=[
            srcs.CASTS,
            srcs.VOTES,
            srcs.TRANSACTIONS])
        address_key = 'orgAddress'
    else:
        raise ValueError(f"Incorrect metric: {metric}")

    return stg, requester, address_key

metrics_dict: Dict[int, Tuple[IMetricStrategy, CacheRequester, str]] = {}

def get_dao(ids: List[str], metric: int) -> MetricDao:
    if metric not in metrics_dict:
        metrics_dict[metric] = _metricsDefault(metric)

    stg, requester, address_key = metrics_dict[metric]
    return MetricDao(ids=ids, strategy=stg, requester=requester, address_key=address_key)



"""
   Descp: Strategy pattern for calculate votes/stakes outcome.

   Created on: 30-mar-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui 
        <f.r.youssef@hotmail.com>
"""

from typing import List
import pandas as pd

from dao_analyzer.apps.common.data_access.daos.metric.strategy import IMetricStrategy
from dao_analyzer.apps.common.business.transfers.stacked_serie import StackedSerie
from dao_analyzer.apps.common.business.transfers.serie import Serie
import dao_analyzer.apps.common.data_access.pandas_utils as pd_utl


VOTES = 0
STAKES = 1


class StTotalVSOption(IMetricStrategy):
    __DF_DATE = 'createdAt'
    __DF_OUTCOME = 'outcome'
    __DF_IS_POSITIVE = 'isPositive'
    __DF_COUNT = 'count'
    __DF_COLS = [__DF_DATE, __DF_IS_POSITIVE]
    __DF_INI_COLS = [__DF_DATE, __DF_OUTCOME]


    def __init__(self, m_type: int):
        self.__type: int = m_type


    def clean_df(self, df: pd.DataFrame) -> pd.DataFrame:
        dff: pd.DataFrame = df
        dff.loc[:, self.__DF_INI_COLS] = dff[self.__DF_INI_COLS]
        return dff


    def process_data(self, df: pd.DataFrame) -> StackedSerie:
        if pd_utl.is_an_empty_df(df):
            return StackedSerie()

        df = self.clean_df(df=df)
        df = self.__transform_df(df=df)

        # join by dates
        df = pd_utl.count_cols_repetitions(df, self.__DF_COLS, self.__DF_COUNT)

        # generates a time serie
        idx = pd_utl.get_monthly_serie_from_df(df, self.__DF_DATE)
        dff = pd_utl.get_df_from_lists([idx, 0], [self.__DF_DATE, self.__DF_COUNT])
        dff = pd_utl.datetime_to_date(dff, self.__DF_DATE)

        serie: Serie = Serie(x=dff[self.__DF_DATE].tolist())
        positives: List = self.__get_outcome(True, df, dff)
        negatives: List = self.__get_outcome(False, df, dff)

        metric: StackedSerie = StackedSerie(
            serie = serie, 
            y_stack = [negatives, positives])

        return metric


    def __get_outcome(self, is_positive: bool, df: pd.DataFrame, dff: pd.DataFrame) -> List:
        d3f: pd.DataFrame = pd_utl.filter_by_col_value(
            df=df, 
            col=self.__DF_IS_POSITIVE, 
            value=is_positive,
            filters=[pd_utl.EQ])

        d3f = d3f.drop(columns=[self.__DF_IS_POSITIVE])
        d3f = pd.concat([d3f, dff], ignore_index=True)
        d3f = d3f.drop_duplicates(subset=self.__DF_DATE, keep="first")
        d3f = d3f.sort_values(self.__DF_DATE)

        return d3f[self.__DF_COUNT].tolist()

    
    def __transform_df(self, df: pd.DataFrame) -> pd.DataFrame:
        dff: pd.DataFrame = df
        outcomes: List[str] = dff[self.__DF_OUTCOME].tolist()

        outcomes: List[bool] = list(map(lambda o: True if o == 'Pass' else False, outcomes))
        dff[self.__DF_IS_POSITIVE] = outcomes

        # takes just the month
        dff = pd_utl.unix_to_date(dff, self.__DF_DATE)
        dff = pd_utl.transform_to_monthly_date(dff, self.__DF_DATE)

        return dff



class Matomo:
    def __init__(self, matomo_url=None, site_id=None):
        self.matomo_url = matomo_url
        self.site_id = site_id

        if not matomo_url:
            raise ValueError("matomo_url has to be set")
        if type(site_id) != int:
            raise ValueError("site_id has to be an integer")

    def get_js(self) -> str:
        return f"""
            var _paq = window._paq = window._paq || [];
            /* tracker methods like "setCustomDimension" should be called before "trackPageView" */
            _paq.push(['trackPageView']);
            _paq.push(['enableLinkTracking']);
            (function() {{
                var u="{self.matomo_url}";
                _paq.push(['setTrackerUrl', u+'matomo.php']);
                _paq.push(['setSiteId', '{self.site_id}']);
                var d=document, g=d.createElement('script'), s=d.getElementsByTagName('script')[0];
                g.async=true; g.src=u+'matomo.js'; s.parentNode.insertBefore(g,s);
            }})();
            """

    def get_script(self) -> str:
        return f"""
            <!-- Matomo -->
            <script>
            {self.get_js()}
            </script>
            <!-- End Matomo Code -->
        """


"""
   Descp: Organization list transfers.

   Created on: 3-mar-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui 
        <f.r.youssef@hotmail.com>
"""
from typing import List, Dict, Union, Set

from .organization import Organization
from .organization_filter import Filter, NetworkFilters, SIMPLE_FILTERS, NetworkRadioButton, OrganizationFilterGroup

from dao_analyzer.apps.common.resources.strings import TEXT

class OrganizationList(list):
    ALL_ORGS_ID = '-1'

    def __init__(self, orgs: Union[List[Organization], List[dict]] = []) -> None:
        # Convert every item to Organization Type
        orgs = list(orgs)
        if len(orgs) > 0:
            if isinstance(orgs[0], Organization):
                super().__init__(orgs)
            else:
                super().__init__(map(Organization.from_json, orgs))
    
    @staticmethod
    def from_json(list: list) -> 'OrganizationList':
        return OrganizationList([] if list is None else list)

    def add_organization(self, org: Organization) -> None:
        if org:
            self.__orgs.append(org)

    def get_organizations(self) -> List[Organization]:
        return list(self)

    def get_all_orgs_dict(self) -> Dict[str, str]:
        return {'value': self.ALL_ORGS_ID, 'label': TEXT['all_orgs']}

    def get_dict_representation(self) -> List[Dict[str, str]]:
        if len(self) == 0:
            return []

        result: List[Dict[str, str]] = [x.get_dict_representation() for x in sorted(self)]

        # Append all_orgs as the first one
        result = [self.get_all_orgs_dict()] + result
        return result

    @classmethod
    def from_dict_representation(cls, dict_repr: List[Dict[str,str]]) -> 'OrganizationList':
        return OrganizationList([Organization(o['value']) for o in dict_repr])

    @classmethod
    def is_all_orgs(cls, o_id: str):
        return o_id == cls.ALL_ORGS_ID


    def is_empty(self) -> bool:
        return len(self) == 0


    def get_size(self) -> int:
        return len(self)


    def get_ids_from_id(self, o_id: str) -> List[str]:
        """
        Gets a list of ids from a o_id attr.
        If o_id is equals to 'all orgs' id then returns a list with all the orgs id.
        If not returns a list with o_id as unique element of the list.
        """
        if self.is_empty():
            return list()

        if not self.is_all_orgs(o_id):
            return [o_id]

        return [x.get_id() for x in self]

    def get_networks(self) -> Set[str]:
        return set((x.get_network().lower() for x in self))

    @staticmethod
    def get_filters(values=None, only_enabled=False, force_disabled=False) -> List[Filter]:
        filters = [f() for f in SIMPLE_FILTERS]

        if values is not None:
            for f in filters:
                f.enabled = f.id in values

        if force_disabled:
            for f in filters:
                f.enabled = False

        if only_enabled:
            filters = [f for f in filters if f.enabled]

        return filters

    @staticmethod
    def get_filter_group(*args, **kwargs) -> OrganizationFilterGroup:
        return OrganizationFilterGroup(OrganizationList.get_filters(*args, **kwargs))

    @staticmethod
    def get_network_filters_for(networks, network_values=None, only_enabled=False, force_disabled=False) -> NetworkFilters:
        nf = NetworkFilters(networks)

        if network_values is not None:
            for f in nf._filters:
                f.enabled = f.id in network_values

        if force_disabled:
            for f in nf._filters:
                f.enabled = False

        if only_enabled:
            nf._filters = [f for f in nf._filters if f.enabled]

        return nf
    
    def get_network_radio(self, network_value=None, **kwargs) -> NetworkRadioButton:
        return NetworkRadioButton(self.get_networks(), network_value)
    
    def get_network_filters(self, network_values=None, only_enabled=False, force_disabled=False) -> NetworkFilters:
        return self.get_network_filters_for(self.get_networks(), network_values=network_values, only_enabled=only_enabled, force_disabled=force_disabled)

    def filter(self, values=None, network_value=None, **kwargs) -> 'OrganizationList':
        """ Returns a new OrganizationList with only filtered items """
        filtered = filter(self.get_filter_group(values, **kwargs).pred, self)
        filtered = filter(self.get_network_radio(network_value, **kwargs).pred, filtered)
        return OrganizationList(filtered)


"""
   Descp: This is a dao (data access object) of the platform.
    It's used to drive the data trasformation from datawarehouse to transfer obj.

   Created on: 01-jun-2022

   Copyright 2022 David Davó Laviña
        <ddavo@ucm.es>
"""
import abc
import pandas as pd

from datetime import datetime

from dao_analyzer.apps.common.business.transfers import Platform
from dao_analyzer.apps.common.business.transfers.organization.organization_list import OrganizationList
from dao_analyzer.apps.common.data_access.requesters.irequester import IRequester

class PlatformDAO(metaclass=abc.ABCMeta):
    def __init__(self, requester: IRequester):
        self._requester = requester

    def get_last_update_str(self) -> datetime:
        return self._requester.get_last_update_str()

    @staticmethod
    def _NaTtoNone(d) -> datetime:
        return None if pd.isna(d) else d

    @abc.abstractmethod
    def get_platform(self, orglist: OrganizationList = None) -> Platform:
        """Gets the platform

        Args:
            orglist (OrganizationList, optional): If not None, limits the organizations used to get the platform info. Defaults to None.

        Returns:
            Platform: The object with the platform information
        """
        raise NotImplementedError
    
    @abc.abstractmethod
    def get_organization_list(self) -> OrganizationList:
        """Gets the organization list

        Returns:
            OrganizationList: The organization list with all its attributes
        """
        raise NotImplementedError



from setuptools import setup

def version_dev(version):
    from setuptools_scm.version import get_local_node_and_date

    if version.dirty:
        return get_local_node_and_date(version)
    else:
        return ""

def main():
    setup(
        # Fix in case we need to build sdist instead
        use_scm_version={
            'local_scheme': version_dev,
            'write_to': 'dao_analyzer/_version.py',
        },
        setup_requires=['setuptools_scm']
    )

if __name__ == "__main__":
    main()


"""
   Descp: This is a dao (data access object) of the organization.
    It's used to drive the data trasformation from datawarehouse to transfer obj.
   Created on: 23-may-2022
   Copyright 2022-2022 David Davó Laviña
        <david@ddavo.me>
"""
import pandas as pd

from dao_analyzer.apps.common.business.transfers import Organization, OrganizationList
from dao_analyzer.apps.common.business.transfers.organization.participation_stats import MembersCreatedProposalsStat, MembersEverVotedStat
from dao_analyzer.apps.common.business.transfers.organization.platform import Platform
from dao_analyzer.apps.common.data_access.daos.platform_dao import PlatformDAO
from dao_analyzer.apps.common.data_access.requesters import CacheRequester

from .metric import srcs

class DaostackDAO(PlatformDAO):
    __DF_IDX = ['network', 'dao']

    __DF_DATE = 'createdAt'

    __DF_CLEAN_COLS = __DF_IDX + [__DF_DATE]

    __DF_PROPOSER = 'proposer'
    __DF_PROP_COLS = __DF_IDX + [__DF_PROPOSER]
    __DF_MEMBER = 'address'
    __DF_MINT_COLS = __DF_IDX + [__DF_MEMBER, __DF_DATE]
    __DF_VOTER = 'voter'
    __DF_VOTES_COLS = __DF_IDX + [__DF_VOTER]

    def __init__(self):
        super().__init__(CacheRequester(srcs=[
            srcs.DAOS,
        ]))

        self._activityRequester = CacheRequester(srcs=[
            srcs.PROPOSALS,
            srcs.VOTES,
            srcs.STAKES,
        ])

        self._mintsCollector = CacheRequester(srcs=[
            srcs.REP_MINTS,
        ])

        self._votesRequester = CacheRequester(srcs=[
            srcs.VOTES,
        ])

        self._propRequester = CacheRequester(srcs=[
            srcs.PROPOSALS,
        ])

    def _get_daos(self) -> pd.DataFrame:
        return self._requester.request().set_index(self.__DF_IDX, drop=False)

    def _get_members(self) -> pd.DataFrame:
        """ Returns all members who, at one time, had some reputation """
        mints = self._mintsCollector.request()
        mints = mints[self.__DF_MINT_COLS]
        mints[self.__DF_DATE] = pd.to_datetime(mints[self.__DF_DATE], unit='s')

        return mints

    def _get_activity(self) -> pd.DataFrame:
        activity: pd.DataFrame = self._activityRequester.request()
        activity = activity[self.__DF_CLEAN_COLS]
        activity[self.__DF_DATE] = pd.to_datetime(activity[self.__DF_DATE], unit='s')
       
        return activity 

    def _get_prop(self, members: pd.DataFrame = None) -> pd.DataFrame:
        prop = self._propRequester.request()
        prop = prop[self.__DF_PROP_COLS]

        # only proposers that are members
        if members is not None:
            prop = prop.merge(members,
                left_on=self.__DF_IDX+[self.__DF_PROPOSER],
                right_on=self.__DF_IDX+[self.__DF_MEMBER],
            )[self.__DF_PROP_COLS]

        return prop

    def _get_votes(self, members: pd.DataFrame = None) -> pd.DataFrame:
        votes = self._votesRequester.request()
        
        # only voters that are members
        if members is not None:
            votes = votes.merge(members, 
                left_on=self.__DF_IDX+[self.__DF_VOTER],
                right_on=self.__DF_IDX+[self.__DF_MEMBER],
            )[self.__DF_VOTES_COLS]

        return votes

    def get_platform(self, orglist: OrganizationList = None) -> Platform:
        daos: pd.DataFrame = self._requester.request()
        
        members = self._get_members()

        if orglist:
            ids = { o.get_id() for o in orglist }

            daos = daos[daos['dao'].isin(ids)]
            members = members[members['dao'].isin(ids)]
        
        # The creation date is when the first member joined
        creation_date = members[self.__DF_DATE].min()

        mcp_pct = self._get_prop(members)[self.__DF_PROPOSER].nunique() / members[self.__DF_MEMBER].nunique()
        mvt_pct = self._get_votes(members)[self.__DF_VOTER].nunique() / members[self.__DF_MEMBER].nunique()

        return Platform(
            name = 'DAOstack',
            creation_date = creation_date,
            networks = list(daos['network'].unique()),
            participation_stats = [
                MembersCreatedProposalsStat(mcp_pct),
                MembersEverVotedStat(mvt_pct),
            ]
        )

    def get_organization_list(self) -> OrganizationList:
        df = self._get_daos()

        dfgb = self._get_activity().groupby(self.__DF_IDX)[self.__DF_DATE]
        df['first_activity'] = dfgb.min()
        df['last_activity'] = dfgb.max()

        # Getting the participation
        members = self._get_members()

        gbp = self._get_prop(members).groupby(self.__DF_IDX)
        gvt = self._get_votes(members).groupby(self.__DF_IDX)
        gbm = members.groupby(self.__DF_IDX)

        df['mcp_pct'] = gbp[self.__DF_PROPOSER].nunique() / gbm[self.__DF_MEMBER].nunique()
        df['mvt_pct'] = gvt[self.__DF_VOTER].nunique() / gbm[self.__DF_MEMBER].nunique()

        l: OrganizationList = OrganizationList()

        for _, org in df.iterrows():
            l.append(Organization(
                network = org['network'],
                o_id = org['dao'],
                name = org['name'],
                creation_date = None, # We don't know how to get it
                first_activity = self._NaTtoNone(org['first_activity']),
                last_activity = self._NaTtoNone(org['last_activity']),
                participation_stats = [
                    MembersCreatedProposalsStat(org['mcp_pct']),
                    MembersEverVotedStat(org['mvt_pct']),
                ],
            ))

        return l



"""
   Descp: This class is used to adapt the vote and stake type
          metric to its visual representation.

   Created on: 14-sep-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

from typing import Dict
from dao_analyzer.apps.common.business.transfers.organization.organization_list import OrganizationList

import dao_analyzer.apps.common.resources.colors as Color
from dao_analyzer.apps.daostack.resources.strings import TEXT
from dao_analyzer.apps.daostack.business.metric_adapter.metric_adapter import MetricAdapter
from dao_analyzer.apps.common.business.transfers.stacked_serie import StackedSerie
import dao_analyzer.apps.daostack.data_access.daos.metric.\
    metric_dao_factory as s_factory

class VoteType(MetricAdapter):
    VOTE: int = 0
    STAKE: int = 1

    def __init__(self, metric_id: int, mtype: int) -> None:
        super().__init__(metric_id)
        self.__mtype: int = self.__get_type(mtype)


    def __get_type(self, mtype: int) -> int:
        t: int = mtype
        if t not in [VoteType.VOTE, VoteType.STAKE]:
            t = VoteType.VOTE
        return t

    
    def get_plot_data(self, o_id: str, organizations: OrganizationList) -> Dict:
        """
        Returns the metric data in a Dict using o_id param.
        """
        dao = s_factory.get_dao(
            ids=organizations.get_ids_from_id(o_id),
            metric=super().metric_id
        )
        metric: StackedSerie = dao.get_metric()

        last_value: int = metric.get_last_value(0) + metric.get_last_value(1)

        return {
            'type1': {
                'y': metric.get_i_stack(0),
                'color': Color.LIGHT_RED,
                'name': TEXT['votes_against'] if self.__mtype is self.VOTE
                    else TEXT['downstakes'],
            },
            'type2': {
                'y': metric.get_i_stack(1),
                'color': Color.LIGHT_GREEN,
                'name': TEXT['votes_for']if self.__mtype is self.VOTE
                    else TEXT['upstakes'],
            },
            'common': {
                'x': metric.get_serie(),
                'type': 'date',
                'x_format': super().DATE_FORMAT,
                'ordered_keys': ['type1', 'type2'],
            },
            'last_serie_elem': metric.get_last_serie_elem(),
            'last_value': last_value,
            'diff': metric.get_diff_last_values(add_stacks=True),
            'diff_rel': metric.get_rel_last_values(add_stacks=True),
        }



"""
   Descp: Common pane which wraps chart layout and other components.

   Created on: 09-sep-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

from dash import html
from dash import dcc
from typing import List, Dict, Union

from dao_analyzer.apps.common.presentation.charts.layout.ilayout import ILayout
from dao_analyzer.apps.common.resources.strings import TEXT
from dao_analyzer.apps.common.presentation.charts.layout.figure.figure import Figure
from dao_analyzer.apps.common.presentation.charts.layout.chart_pane_configuration \
    import ChartPaneConfiguration


class ChartPaneLayout(ILayout):
    PANE_ID: int = 1
    SUFFIX_ID_CHART: str = '-chart'
    SUFFIX_ID_SUBTITLE1: str = '-subtitle1'
    SUFFIX_ID_SUBTITLE2: str = '-subtitle2'


    def __init__(self, title: str, css_id: str, figure: Figure, css_classes: Union[List[str], str] = []) -> None:
        self.__title: str = title
        self.__css_id: str = css_id
        self.__figure: Figure = figure
        if isinstance(css_classes, str):
            self.__css_classes: List[str] = css_classes.split(' ')
        else:
            self.__css_classes: List[str] = css_classes
        self.__configuration: ChartPaneConfiguration = ChartPaneConfiguration(
            figure.configuration)

    def get_layout(self) -> html.Div:
        """
        Returns a pane with all the components initialized .
        """
        figure: Dict = self.__figure.get_empty_figure()
        subtitle1: str = TEXT['default_amount']
        subtitle2: str = TEXT['no_data_selected']
        children = self._get_children(subtitle1, subtitle2, figure)

        return html.Div(
            children=dcc.Loading(
                type="circle",
                color=self.configuration.color,
                children=html.Div(
                    children=children,
                    id=self.__css_id,
                    className='d-flex flex-column'
            )),  
            className=f'pane {self.configuration.css_border} col {" ".join(self.__css_classes)}'
        )


    def fill_child(self, plot_data: Dict = None) -> List:
        if plot_data:
            figure = self.__figure.get_figure(plot_data=plot_data)
        else:
            figure = self.__figure.get_empty_figure()
        subtitle1: str = TEXT['default_amount']
        subtitle2: str = TEXT['no_data_selected']

        if plot_data and self.configuration.show_subtitles:
            subtitle1 = TEXT['graph_amount'].format(
                plot_data['last_serie_elem'], 
                plot_data['last_value']
            )
            subtitle2 = TEXT['graph_subtitle'].format(plot_data['diff_rel'])

        return self._get_children(subtitle1, subtitle2, figure)


    def _get_children(self, subtitle1, subtitle2, figure) -> List:
        hide: str = '' if self.configuration.show_subtitles else 'hide'

        return [
            html.Div(children=[
                html.Span(
                    self.__title,
                    className='graph-pane-title'
                ),
                html.Span(
                    subtitle2, 
                    id=f'{self.__css_id}{self.SUFFIX_ID_SUBTITLE2}',
                    className=f'{hide}'
                ),
            ], className='d-flex flex-column chart-header'),
            dcc.Graph(
                id=f'{self.__css_id}{self.SUFFIX_ID_CHART}',
                figure=figure,
                className='chart',
            ),
        ]


    @property
    def configuration(self) -> ChartPaneConfiguration:
        return self.__configuration


    @property
    def figure(self) -> Figure:
        return self.__figure



"""
   Descp: Util for pandas interactions.

   Created on: 16-mar-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

from datetime import date
from typing import List, Any, Dict, Tuple
import pandas as pd
from pandas import DataFrame
from pandas.tseries.offsets import DateOffset


EQ = 0
NEQ = 1
GT = 2
GTE = 3
LT = 4
LTE = 5


def filter_by_col_value(df: DataFrame, col: str, value: Any, 
filters: List[int]) -> DataFrame:
    """
    Filters rows by a column value with each filter on filters list.
    Arguments:
        * df = The data frame to filter.
        * col = column name which is selected to filter by.
        * value = The value which the column must ajust.
        * filters = A list of filter to apply on the data frame.
    Return:
        A filtered data frame copy.
    """
    dff: DataFrame = df
    for f in filters:
        if f == EQ:
            dff = dff.loc[dff[col] == value]
        elif f == NEQ:
            dff = dff.loc[dff[col] != value]
        elif f == GT:
            dff = dff.loc[dff[col] > value]
        elif f == GTE:
            dff = dff.loc[dff[col] >= value]
        elif f == LT:
            dff = dff.loc[dff[col] < value]
        elif f == LTE:
            dff = dff.loc[dff[col] <= value]
    
    return dff

# TODO: Delete this (stop using it)
def append_rows(df: DataFrame, row: List) -> DataFrame:
    dff = pd.DataFrame([pd.Series(row, index=df.columns)])
    return pd.concat([df, dff], join='outer', ignore_index=True)


def get_empty_data_frame(columns: List[str] = None) -> DataFrame:
    cols = columns if columns else list()
    return DataFrame(columns=cols)


def get_df_from_lists(rows: List[List], columns: List[str]) -> DataFrame:
    if not len(rows) == len(columns):
        raise Exception

    rows = rows if rows else list()
    columns = columns if columns else list()

    di_df: Dict = dict()
    for i, c in enumerate(columns):
        di_df[c] = rows[i]

    return DataFrame(di_df).infer_objects()


def is_an_empty_df(df: DataFrame) -> bool:
    return df.shape[0] == 0


def unix_to_date(df: DataFrame, col: str) -> DataFrame:
    dff: DataFrame = df.copy()
    dff.loc[:, col] = pd.to_datetime(dff.loc[:, col], unit='s').dt.date
    return dff


def transform_to_monthly_date(df: DataFrame, col: str) -> DataFrame:
    dff: DataFrame = df.copy()
    dff.loc[:, col] = dff[col].apply(lambda d: d.replace(day=1))
    return dff


def datetime_to_date(df: DataFrame, col: str) -> DataFrame:
    dff: DataFrame = df
    dff.loc[:, col] = dff[col].dt.date
    return dff


def count_cols_repetitions(df: DataFrame, cols: List[str], new_col: str)\
-> DataFrame:
    dff: DataFrame = df
    dff = dff.groupby(cols).size().reset_index(name=new_col)
    return dff


def sum_cols_repetitions(df: DataFrame, cols: List[str], new_col: str)\
-> DataFrame:
    dff: DataFrame = df
    dff = dff.groupby(cols).sum().reset_index()
    return dff


def get_monthly_serie_from_df(df: DataFrame, date_col: str, start: date = None) -> pd.DatetimeIndex:
    today = date.today().replace(day=1)
    start = df[date_col].min() if not start else start
    return pd.date_range(start=start, end=today, freq=DateOffset(months=1))


def drop_duplicate_date_rows(df: DataFrame, dff: DataFrame, date_col: str) -> None:
    """
    Removes all duplicated date-rows from dff considering the df's ones.
    """
    mask = dff[date_col].isin(df[date_col])
    dff = dff.drop(dff[mask].index)

    return dff

def top_rest_daos(df: DataFrame, idx, value_col, top_pct: float=0.50) -> Tuple[DataFrame, DataFrame]:
    """
    Calculates the top DAOs of each network
    """
    dfsum = df.groupby([idx, 'network']).sum() # Calculating value of every DAO
    dfsum = dfsum.sort_values(['network', value_col]) # Sorting by network>value because group doesnt have sort_values
    gbsum = dfsum.groupby('network')
    cumval = gbsum[value_col].cumsum() / gbsum[value_col].sum() # Calculating the cummulative value

    topmask = cumval > top_pct

    # Getting the top/rest subsets
    top = dfsum[topmask]
    rest = dfsum[~topmask]

    # Applying them to the original dfs (keeping all data)
    top = df.loc[top.index]
    rest = df.loc[rest.index]

    return top, rest



"""
   Descp: Installed apps test.

   Created on: 20-oct-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

import unittest
from typing import List, Dict
import pandas as pd

from dao_analyzer.apps.aragon.data_access.daos.metric.strategy.\
    st_installed_apps import StInstalledApps

from dao_analyzer.apps.common.business.transfers.stacked_serie import StackedSerie


class StInstalledAppsTest(unittest.TestCase):

    def test_repeated_names(self):

        in_df: pd.DataFrame = pd.DataFrame([
            {'trash': 'trash', 'id': '1', 'repoName': '1'},
            {'trash': 'trash', 'id': '2', 'repoName': '2'},
            {'trash': 'trash', 'id': '3', 'repoName': '3'},
            {'trash': 'trash', 'id': '4', 'repoName': '1'},
        ])
        strategy: StInstalledApps = StInstalledApps()
        result: StackedSerie = strategy.process_data(df=in_df)
        out: List[int] = [2, 1, 1]

        self.assertListEqual(out, result.get_i_stack(i_stack=0))


    def test_repeated_non_names(self):

        in_df: pd.DataFrame = pd.DataFrame([
            {'trash': 'trash', 'id': '1', 'repoName': '1'},
            {'trash': 'trash', 'id': '2', 'repoName': '2'},
            {'trash': 'trash', 'id': '3', 'repoName': None},
            {'trash': 'trash', 'id': '4', 'repoName': None},
        ])
        strategy: StInstalledApps = StInstalledApps()
        result: StackedSerie = strategy.process_data(df=in_df)
        out: List[int] = [1, 1]

        self.assertListEqual(out, result.get_i_stack(i_stack=0))


    def test_repetitions(self):

        in_df: pd.DataFrame = pd.DataFrame([
            {'trash': 'trash', 'id': '1', 'repoName': '1'},
            {'trash': 'trash', 'id': '1', 'repoName': '1'},
            {'trash': 'trash', 'id': '1', 'repoName': '1'},
        ])
        strategy: StInstalledApps = StInstalledApps()
        result: StackedSerie = strategy.process_data(df=in_df)
        out: List[int] = [1]

        self.assertListEqual(out, result.get_i_stack(i_stack=0))


    def test_others(self):
        max_elems: int = 15

        in_data: List[Dict[str, str]] = [
            {'id': '1', 'repoName': '1'},
            {'id': '2', 'repoName': '1'},
            {'id': '3', 'repoName': '1'},
        ] + \
        [{'id': f'{i}', 'repoName': f'{i}'} for i in range(4, max_elems+6)]

        in_df: pd.DataFrame = pd.DataFrame(in_data)
        strategy: StInstalledApps = StInstalledApps()
        result: StackedSerie = strategy.process_data(df=in_df)

        out: List[int] = [3] + ([1] * (max_elems-1)) + [3] 

        self.assertListEqual(out, result.get_i_stack(i_stack=0))


if __name__ == "__main__":
    unittest.main()



"""
    Descp: Aragon Runner and Collectors

    Created on: 02-nov-2021

    Copyright 2021 David Davó
        <david@ddavo.me>
"""
from typing import List

import pkgutil
from gql.dsl import DSLField
import pandas as pd
import numpy as np
import json

from ..common.cryptocompare import CCPricesCollector
from ..common import ENDPOINTS, Collector
from ..common.graphql import GraphQLCollector, GraphQLRunner
from ..common.blockscout import BlockscoutBallancesCollector

class AppsCollector(GraphQLCollector):
    def __init__(self, runner, network: str):
        super().__init__('apps', runner, endpoint=ENDPOINTS[network]['aragon'], network=network)

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.apps(**kwargs).select(
            ds.App.id,
            ds.App.isForwarder,
            ds.App.isUpgradeable,
            ds.App.repoName,
            ds.App.repoAddress,
            ds.App.organization.select(ds.Organization.id)
        )

class BalancesCollector(BlockscoutBallancesCollector):
    def __init__(self, runner, base, network: str):
        super().__init__(runner, addr_key='recoveryVault', base=base, network=network)

class CastsCollector(GraphQLCollector):
    def __init__(self, runner, network: str):
        super().__init__('casts', runner, endpoint=ENDPOINTS[network]['aragon_voting'], network=network, pbar_enabled=False)

        @self.postprocessor
        def changeColumnNames(df: pd.DataFrame) -> pd.DataFrame:
            df = df.rename(columns={
                'voterId':'voter', 
                'voteAppAddress':'appAddress',
                'voteOrgAddress':'orgAddress'})
            return df

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.casts(**kwargs).select(
            ds.Cast.id,
            ds.Cast.vote.select(ds.Vote.id),
            ds.Cast.voter.select(ds.Voter.id),
            ds.Cast.supports,
            ds.Cast.stake,
            ds.Cast.createdAt,
            ds.Cast.vote.select(
                ds.Vote.orgAddress,
                ds.Vote.appAddress
            )
        )

class OrganizationsCollector(GraphQLCollector):
    DAO_NAMES=pkgutil.get_data('cache_scripts.aragon', 'dao_names.json')

    def __init__(self, runner, network: str):
        super().__init__('organizations', runner, endpoint=ENDPOINTS[network]['aragon'], network=network)

        @self.postprocessor
        def set_dead_recoveryVault(df: pd.DataFrame) -> pd.DataFrame:
            df['recoveryVault'] = df['recoveryVault'].replace(r'^0x0+$', np.NaN, regex=True)
            return df

        @self.postprocessor
        def apply_names(df: pd.DataFrame) -> pd.DataFrame:
            names_dict = json.loads(self.DAO_NAMES)

            if self.network not in names_dict.keys() or \
            not names_dict[self.network] or \
            df.empty:
                return df

            names_df = pd.json_normalize(names_dict[self.network])
            names_df['id'] = names_df['address'].str.lower()
            names_df['name'] = names_df['name'].fillna(names_df['domain'])
            names_df = names_df[['id', 'name']]
            df = df.merge(names_df, on='id', how='left')

            return df

        @self.postprocessor
        def copy_id(df: pd.DataFrame) -> pd.DataFrame:
            df['orgAddress'] = df['id']
            return df

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.organizations(**kwargs).select(
            ds.Organization.id,
            ds.Organization.createdAt,
            ds.Organization.recoveryVault
        )

class MiniMeTokensCollector(GraphQLCollector):
    def __init__(self, runner, network: str):
        super().__init__('miniMeTokens', runner, endpoint=ENDPOINTS[network]['aragon_tokens'], network=network, pbar_enabled=False)

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.miniMeTokens(**kwargs).select(
            ds.MiniMeToken.id,
            ds.MiniMeToken.address,
            ds.MiniMeToken.totalSupply,
            ds.MiniMeToken.transferable,
            ds.MiniMeToken.name,
            ds.MiniMeToken.symbol,
            ds.MiniMeToken.orgAddress,
            ds.MiniMeToken.appAddress,
            ds.MiniMeToken.lastUpdateAt
        )

class TokenHoldersCollector(GraphQLCollector):
    def __init__(self, runner: GraphQLRunner, network: str):
        super().__init__('tokenHolders', runner, endpoint=ENDPOINTS[network]['aragon_tokens'], network=network)

        @self.postprocessor
        def add_minitokens(df: pd.DataFrame) -> pd.DataFrame:
            tokens = runner.filterCollector(name='miniMeTokens', network=network).df
            tokens = tokens.rename(columns={'address':'tokenAddress', 'orgAddress':'organizationAddress'})
            return df.merge(tokens[['tokenAddress', 'organizationAddress']], on='tokenAddress', how='left')
            
    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.tokenHolders(**kwargs).select(
            ds.TokenHolder.id,
            ds.TokenHolder.address,
            ds.TokenHolder.tokenAddress,
            ds.TokenHolder.lastUpdateAt,
            ds.TokenHolder.balance
        )

class TokenPricesCollector(CCPricesCollector):
    pass

class ReposCollector(GraphQLCollector):
    def __init__(self, runner, network: str):
        super().__init__('repos', runner, network=network, endpoint=ENDPOINTS[network]['aragon'])

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.repos(**kwargs).select(
            ds.Repo.id,
            ds.Repo.address,
            ds.Repo.name,
            ds.Repo.node,
            ds.Repo.appCount
        )

class TransactionsCollector(GraphQLCollector):
    def __init__(self, runner, network: str):
        super().__init__('transactions', runner, network=network, endpoint=ENDPOINTS[network]['aragon_finance'])

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.transactions(**kwargs).select(
            ds.Transaction.id,
            ds.Transaction.orgAddress,
            ds.Transaction.appAddress,
            ds.Transaction.token,
            ds.Transaction.entity,
            ds.Transaction.isIncoming,
            ds.Transaction.amount,
            ds.Transaction.date,
            ds.Transaction.reference
        )

class VotesCollector(GraphQLCollector):
    def __init__(self, runner, network: str):
        super().__init__('votes', runner, network=network, endpoint=ENDPOINTS[network]['aragon_voting'])

    def query(self, **kwargs) -> DSLField:
        ds = self.schema
        return ds.Query.votes(**kwargs).select(
            ds.Vote.id,
            ds.Vote.orgAddress,
            ds.Vote.appAddress,
            ds.Vote.creator,
            ds.Vote.originalCreator,
            ds.Vote.metadata,
            ds.Vote.executed,
            ds.Vote.executedAt,
            ds.Vote.startDate,
            ds.Vote.supportRequiredPct,
            ds.Vote.minAcceptQuorum,
            ds.Vote.yea,
            ds.Vote.nay,
            ds.Vote.voteNum,
            ds.Vote.votingPower
        )

class AragonRunner(GraphQLRunner):
    name: str = 'aragon'

    def __init__(self, dw=None):
        super().__init__(dw)
        self._collectors: List[Collector] = []
        ## TODO: Fix aragon-tokens xdai subgraph and redeploy
        self.networks = ['mainnet']

        for n in self.networks: 
            self._collectors.extend([
                AppsCollector(self, n),
                CastsCollector(self, n),
                MiniMeTokensCollector(self, n),
                ReposCollector(self, n),
                TransactionsCollector(self, n),
                TokenHoldersCollector(self, n),
                VotesCollector(self, n)
            ])
            oc = OrganizationsCollector(self, n)
            bc = BalancesCollector(self, oc, n)
            self._collectors += [oc, bc]
        
        self._collectors.append(CCPricesCollector(self))

    @property
    def collectors(self) -> List[Collector]:
        return self._collectors


import pandas as pd
import numpy as np

from cache_scripts.common.api_requester import CryptoCompareRequester

from .. import config
from .common import Collector, NetworkRunner

import logging

EMPTY_KEY_MSG = \
"""\
Empty CryptoCompare API key. You can obtain one from https://www.cryptocompare.com/cryptopian/api-keys
You can set the API key using --cc-api-key argument or the DAOA_CC_API_KEY env variable.
"""

def cc_postprocessor(df: pd.DataFrame) -> pd.DataFrame:
    ccrequester = CryptoCompareRequester(api_key=config.cc_api_key)

    tokenSymbols = df['symbol'].drop_duplicates()

    # TODO: Get only the ones with available symbols (relaxedValidation=False)
    df_fiat = pd.DataFrame.from_dict(ccrequester.get_symbols_price(tokenSymbols, relaxedValidation=True), orient='index')

    def _apply_values(row):
        if row['symbol'] in df_fiat.index:
            row['usdValue'] = row['balanceFloat'] * df_fiat.loc[row['symbol'], 'USD']
            row['ethValue'] = row['balanceFloat'] * df_fiat.loc[row['symbol'], 'ETH']
            row['eurValue'] = row['balanceFloat'] * df_fiat.loc[row['symbol'], 'EUR']
        else:
            row['usdValue'] = np.NaN
            row['ethValue'] = np.NaN
            row['eurValue'] = np.NaN
        
        return row

    df = df.apply(_apply_values, axis='columns')

    return df

class CCPricesCollector(Collector):
    def __init__(self, runner: NetworkRunner, name: str='tokenPrices'):
        super().__init__(name, runner)
        self.requester = CryptoCompareRequester(api_key=config.cc_api_key)

    def verify(self) -> bool:
        if not self.requester.api_key:
            logging.warning(EMPTY_KEY_MSG)
            return False

        return super().verify()
    
    @property
    def base(self):
        return self.runner.filterCollector(name='tokenBalances')

    def run(self, force=False, block=None):
        tokenSymbols = pd.read_feather(self.base.data_path, columns=['symbol']).drop_duplicates()['symbol']
        # TODO: Get only coins with available info (relaxedValidation=False)

        df = pd.DataFrame.from_dict(self.requester.get_symbols_price(tokenSymbols, relaxedValidation=True), orient='index')
        df.reset_index().to_feather(self.data_path)


"""
   Descp: This file stores the various sources as an enum

   Created on: 13-dec-2021

   Copyright 2021 David Davó
        <david@ddavo.me>
"""

from pathlib import Path
from typing import List

CACHE_PATH: Path = Path('datawarehouse') / 'aragon'
APPS: Path = CACHE_PATH / 'apps.arr'
CASTS: Path = CACHE_PATH / 'casts.arr'
MINI_ME_TOKENS: Path = CACHE_PATH / 'miniMeTokens.arr'
ORGANIZATIONS: Path = CACHE_PATH / 'organizations.arr'
REPOS: Path = CACHE_PATH / 'repos.arr'
TOKEN_BALANCES: Path = CACHE_PATH / 'tokenBalances.arr'
TOKEN_HOLDERS: Path = CACHE_PATH / 'tokenHolders.arr'
TRANSACTIONS: Path = CACHE_PATH / 'transactions.arr'
VOTES: Path = CACHE_PATH / 'votes.arr'
ALL_FILES: List[str] = [
    APPS,
    CASTS,
    MINI_ME_TOKENS,
    ORGANIZATIONS,
    REPOS,
    TOKEN_BALANCES,
    TOKEN_HOLDERS,
    TRANSACTIONS,
    VOTES
]


"""
   Descp: Wraps the figure configuration.

   Created on: 10-sep-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

from typing import Dict, List, Tuple

import dao_analyzer.apps.common.resources.colors as Color

class FigureConfiguration():

    def __init__(self) -> None:
        self.__show_legend: bool = True
        self.__horizontal_lines: List[Tuple[float, str]] = list()
        self.__x_layout_params: List[Dict] = [{}]
        self.__y_layout_params: List[Dict] = [{}]


    def enable_legend(self) -> None:
        self.__show_legend = True


    def disable_legend(self) -> None:
        self.__show_legend = False

    
    def get_legend(self) -> Dict:
        legend: Dict = {'orientation': 'h', 'x': 0, 'y': 1.2}
        return legend if self.__show_legend else {}


    def add_horizontal_line(self, y: float, y_ref: str) -> None:
        """
        Adds a horizontal line in the list.
        Params:
            * y: its value in the y-axis
            * y_ref: it selects which y-axis you want, in the case there
                     is more than one. E.g. y, y2, y3, ...
        """
        self.__horizontal_lines.append((y, y_ref))


    def get_shapes(self) -> List[Dict]:
        shapes: List[Dict] = list()

        for y, y_ref in self.__horizontal_lines:
            shapes.append({
                'type': 'line',
                'xref': 'paper',
                'yref': y_ref,
                'x0': 0,
                'y0': y,
                'x1': 1,
                'y1': y,
                'line':{
                    'color': Color.BLACK,
                    'width': 0.5,
                    'dash':'dot'
                }
            })

        return shapes


    def get_x_axis_layout(self, axis: int = 0) -> Dict:
        return self.__get_axis_layout(args=self.__x_layout_params[axis])


    def get_y_axis_layout(self, axis: int = 0) -> Dict:
        return self.__get_axis_layout(args=self.__y_layout_params[axis]) 


    def __get_axis_layout(self, args: Dict) -> Dict:
        """
        Returns the axis layout using params in args.
        TODO: remove args parameter to simplify.
        """
        axis_l: Dict[str, str] = {
            'type': args['type'] if 'type' in args else '-',
            'ticks': 'outside',
            'ticklen': 5,
            'tickwidth': 2,
            'ticksuffix': args['suffix'] if 'suffix' in args else '',
            'showline': True, 
            'linewidth': 2, 
            'linecolor': Color.BLACK,
            'showgrid': args['grid'] if 'grid' in args else False,
            'gridwidth': 0.5,
            'gridcolor': Color.GRID_COLOR,
            'tickfont_color': Color.TICKFONT_COLOR,
            'automargin': True,
        }

        if 'removemarkers' in args:
            axis_l['ticklen'] = 0
            axis_l['tickwidth'] = 0
        if 'tickvals' in args:
            axis_l['tickvals'] = args['tickvals']
        if 'reverse_range' in args:
            axis_l['autorange'] = 'reversed'
        if 'l_range' in args:
            axis_l['range'] = args['l_range']
        if 'tickformat' in args:
            axis_l['tickformat'] = args['tickformat']
        if 'tickangle' in args:
            axis_l['tickangle'] = 45
        if 'matches' in args:
            axis_l['matches'] = args['matches']
        if 'tickfont_size' in args:
            axis_l['tickfont_size'] = args['tickfont_size']

        return axis_l

    def get_height(self) -> int:
        return 350

    def get_margin(self) -> Dict[str, int]:
        # Margin is calculated automatically
        return {
            'l': 0,
            'r': 0,
            't': 0,
            'b': 0,
        }

    def add_axis(self, x_axis: int, y_axis: int) -> None:
        self.__x_layout_params = [{} for _ in range(x_axis)]
        self.__y_layout_params = [{} for _ in range(y_axis)]


    def add_x_params(self, params: Dict, axis: int = 0) -> None:
        self.__add_params(params=params, axis=self.__x_layout_params[axis])


    def add_y_params(self, params: Dict, axis: int = 0) -> None:
        self.__add_params(params=params, axis=self.__y_layout_params[axis])


    def __add_params(self, params: Dict, axis: Dict) -> None:
        for k, v in params.items():
            axis[k] = v



"""
   Descp: Use this file in order to isolate your app logs

   Created on: 24-feb-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui 
        <f.r.youssef@hotmail.com>
"""

LOGS: dict = {
    'attr_not_init': 'You haven\'t initialized {} attribute\
, you can\'t use {} method directly.',
    'chunks_requested': 'Requested {} chunk(s) in {:.2f} ms',
    'daos_requested': '{} DAO(s) requested in {:.2f} s',
    'graph_error': 'An error has ocurred or no data available.',
    'no_attr_api_req': '\'args\' has to be filled with a query.\n',
    'no_attr_cache_req': '\'args\' has to be filled with a cache path.\n',
    'request_to': 'Requesting to: {}',
    'requested_in': 'Requested in {:.2f} ms',
}


"""
   Descp: Strategy pattern to create active DAO(meaning organizations) metric.
          It'll show the number of active DAOs each month,
          whether just one DAO is selected it will show the month when 
          it was active.

   Created on: 23-oct-2020

   Copyright 2020-2021 Youssef 'FRYoussef' El Faqir El Rhazoui
        <f.r.youssef@hotmail.com>
"""

import pandas as pd

from dao_analyzer.apps.common.data_access.daos.metric.strategy import IMetricStrategy

from dao_analyzer.apps.common.business.transfers.stacked_serie import StackedSerie
from dao_analyzer.apps.common.business.transfers.serie import Serie
import dao_analyzer.apps.common.data_access.pandas_utils as pd_utl


class StActiveOrganization(IMetricStrategy):
    __DF_DATE = 'createdAt'
    __DF_ID = 'dao'
    __DF_COUNT = 'count'
    __DF_COLS = [__DF_DATE, __DF_ID]


    def clean_df(self, df: pd.DataFrame) -> pd.DataFrame:
        dff: pd.DataFrame = df
        dff.loc[:, self.__DF_COLS] = dff[self.__DF_COLS]
        return dff


    def process_data(self, df: pd.DataFrame) -> StackedSerie:
        if pd_utl.is_an_empty_df(df):
            return StackedSerie()

        df = self.clean_df(df=df)

        # takes just the month
        df = pd_utl.unix_to_date(df, self.__DF_DATE)
        df = pd_utl.transform_to_monthly_date(df, self.__DF_DATE)

        # join dates-ids
        df = pd_utl.count_cols_repetitions(df, self.__DF_COLS, self.__DF_COUNT)
        # different users by month
        df = pd_utl.count_cols_repetitions(df, [self.__DF_DATE], self.__DF_COUNT)

        # generates a time series
        idx = pd_utl.get_monthly_serie_from_df(df, self.__DF_DATE)
        dff = pd_utl.get_df_from_lists([idx, 0], [self.__DF_DATE, self.__DF_COUNT])
        dff = pd_utl.datetime_to_date(dff, self.__DF_DATE)

        # joinning all the data in a unique dataframe
        df = pd.concat([df, dff], ignore_index=True)
        df = df.drop_duplicates(subset=self.__DF_DATE, keep="first")
        df = df.sort_values(self.__DF_DATE)

        serie: Serie = Serie(x=df[self.__DF_DATE].tolist())
        metric: StackedSerie = StackedSerie(
            serie = serie, 
            y_stack = [df[self.__DF_COUNT].tolist()])

        return metric
